/*
 * This program decodes Frontier ELITEII pic*.grf files.
 *
 * $Log: seegrf.c $
 * Revision 1.1  1995/05/21  11:55:16  spc
 * Initial revision
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <time.h>
#include <io.h>
#include <dos.h>
#include <graph.h>
#include <conio.h>

static const char RcsId[] = "$Id: seegrf.c 1.1 1995/05/21 11:55:16 spc Exp $" ;

/*
 * structure of the colour table in the .GRF file
 */
typedef  struct
{
   unsigned char  r,g,b ;
}
rgb_st ;

typedef  struct
{
   char  *cmd ;                  /* name of the command */
   int   (*fn)(int, char*[]) ;   /* function to deal with it */
   char  *desc ;                 /* brief description */
}
cmd_opt_st ;

#define FALSE 0
#define TRUE 1

/*
 * dimensions of screen in pixels
 */
#define SCR_WIDTH 320
#define SCR_DEPTH 200

#define SCREEN_LEN 0xfa00

/* 
 * The colour table is always 0x300 bytes long.
 *
 * The image data is coded as follows:
 * Sequences of 2 or more of the same byte value are stored in 2 bytes.
 * Differing byte sequences have an overhead of 1 byte per approx 125 bytes stored.
 *
 * The maximum number of bytes to be coded is 0xfa00 (=64000).
 * If every byte differed from the previous byte, then there would be
 * (0xfa00/125) 512 more bytes required.
 */
#define MAX_FILE_SIZE   (SCREEN_LEN+0x300+512)

/*
 * size of colour table
 */
#define COL_SIZE  (sizeof(rgb_st)*256)

short       last_x = 0, last_y = 0 ;
char __far  *screen ;
long        palette[256] ;

/*
 * options variables
 */
int   frame_delay    = 0 ;
int   loop_forever   = FALSE ;


/*
 * sets up a new screen mode.  Saves the old mode
 * so that it can be restored when we have finished.
 */
short old_mode ;
void screen_mode ( short mode )
{
   extern void fatal ( char *msg ) ;
   short       status ;
   struct      _videoconfig vc;

   _getvideoconfig ( &vc ) ;
   old_mode = vc.mode ;
   status = _setvideomode ( mode ) ;
   if ( status == 0 )
   {
      fatal ( "Your hardware cannot use the correct mode" ) ;
   }
}


/*
 * fatal error occurred, so print a message and exit.
 */
void fatal ( char *msg )
{
   screen_mode ( old_mode ) ;
   fprintf ( stderr, "%s\n", msg ) ;
   exit ( 2 ) ;
}


/*
 * Plots a number of points in a colour.
 * This uses the supplied DOS routines for writing to the screen.
 */
void plot_slow ( short col, int count )
{
   int   i ;
   short old_col ;

   old_col = _setcolor ( col & 0xff ) ;
   for ( i = 0 ; i < count ; i++ )
   {
      _setpixel ( last_x, last_y ) ;
      last_x++ ;
      if ( last_x == SCR_WIDTH ) { last_x = 0 ; last_y++ ; }
      if ( last_y == SCR_DEPTH ) { last_y = 0 ; }
   }
}

/*
 * Plots a number of points in a colour.
 * This writes directly to screen memory.
 */
void plot_fast ( short col, int count )
{
   int i ;

   for ( i = 0 ; i < count ; i++ )
   {
      *screen++ = (char)col ;
   }
}


/*
 * select the plotting function.  By default, we will
 * use the faster method of writing directly to the screen
 * memory.
 */
void (*plot)(short, int) = plot_fast ;


/*
 * define the colour palette for the saved image
 */
void do_colour ( rgb_st *table )
{
   int   i ;
   short status ;

   for ( i = 0 ; i < 256 ; i++ )
   {
      /* convert to palette form */
      palette[i] = (long)table->b << 16 | 
                   (long)table->g << 8  |
                   (long)table->r ;
      table++ ;
   }

   status = _remapallpalette ( palette ) ;
   if ( status == 0 )
   {
      fatal ( "Failed to define palette" ) ;
   }
}


/*
 * data is as follows:
 * if the current byte is -ve, repeat the following byte -(count-1) times
 * if the current byte is +ve, take the next (count+1) bytes literally
 */
void do_rle ( signed char *data, long num_entries )
{
   long     i = 0 ;     /* indexes the data */
   int      count ;     /* current data set length */
   long int sum = 0 ;   /* debug, counts number of expanded bytes */
   short    colour ;    /* color of current pixel */

   while ( ( i < num_entries ) && ( sum <= SCREEN_LEN ) )
   {
      count = (int)data[i++] ;
      if ( count < 0 )
      {
         /* length encoded data */
         count = -(count-1) ;
         colour = (short)data[i++] ;
         plot ( colour, count ) ;
         sum += count ;
      }
      else
      {
         /* literally stored data */
         count++ ;
         for ( ; count > 0 ; count-- )
         {
            plot ( (short)data[i++], 1 ) ;
            sum++ ;
         }
      }
   }
}


/*
 * checks that all values in a palette are in range
 */
int check_palette ( unsigned char *mem )
{
   int i, result = TRUE ;

   for ( i = 0 ; i < COL_SIZE ; i++ )
   {
      result &= *mem++ <= 0x3f ;
   }
   return result ;
}


/*
 * reads and processes a GRF file
 */
void read_file ( char *filename, int fd, long filesize, void *mem )
{
   long  num_read ;

   /* read the colour table information */
   num_read = read ( fd, mem, (unsigned)COL_SIZE ) ;
   if ( num_read != COL_SIZE )
   {
      fprintf ( stderr, "%s is too small for a GRF file\n", filename ) ;
      return ;
   }
   if ( !check_palette(mem) )
   {
      fprintf ( stderr, "%s has bad RGB values\n", filename ) ;
      return ;
   }
   do_colour ( mem ) ;

   /* read and process the encoded image data */
   num_read = read ( fd, mem, (unsigned)(filesize-(long)COL_SIZE) ) ;
   if ( num_read != (int)(filesize-COL_SIZE) )
   {
      fprintf ( stderr, "%s is too small for a GRF file\n", filename ) ;
      return ;
   }
   do_rle ( mem, num_read ) ;
}


/*
 * open the GRF file.
 */
void decode_file ( char *filename )
{
   int   fd ;
   long  filesize ;
   void  *mem ;

   fd = open ( filename, _O_RDONLY | _O_BINARY, 0 ) ;
   filesize = filelength ( fd ) ;
   if ( filesize > MAX_FILE_SIZE )
   {
      fprintf ( stderr, "%s is too big for a GRF file\n", filename ) ;
      close ( fd ) ;
      return ;
   }

   mem = malloc ( (size_t)__max((long)COL_SIZE,(filesize-(long)COL_SIZE)) ) ;

   read_file ( filename, fd, filesize, mem ) ;

   free ( mem ) ;

   close ( fd ) ;
}

/*
 * tests for a key.
 * returns TRUE if there is a key waiting.
 */
int test_for_key ( void )
{
   union _REGS regs ;
   regs.h.ah = 0x0b ;
   _int86 ( 0x21, &regs, &regs ) ;
   return ( regs.h.al == 255 ) ;
}


/*
 * Pauses for a specified number of microseconds,
 * or until a key is pressed.
 */
void sleep( clock_t wait )
{
   clock_t  goal ;
   goal = wait + clock() ;
   while( goal > clock() )
   {
      if ( test_for_key() ) break ; 
   }
}


/*
 * functions to action the command line options
 */
int opt_slow_fn ( int argc, char *argv[] )
{
   plot = plot_slow ;
   return argc + 1 ;
}

int opt_loop_fn ( int argc, char *argv[] )
{
   loop_forever = TRUE ;
   return argc + 1 ;
}

int opt_delay_fn ( int argc, char *argv[] )
{
   if ( sscanf ( argv[argc+1], "%i", &frame_delay ) != 1 )
   {
      /* failed to decode the expected integer, force it to be zero */
      frame_delay = 0 ;
   }
   return argc + 2 ;
}

int opt_version_fn ( int argc, char *argv[] )
{
   printf ( "%s\n", RcsId ) ;
   return argc + 1 ;
}


cmd_opt_st  cmd_opts[] =
{
   { "-slow",  opt_slow_fn,   "    Use slow screen write routine" },
   { "-loop",  opt_loop_fn,   "    Continuous repetetive display" },
   { "-delay", opt_delay_fn,  "nn  Wait nn seconds between frames" },
   { "-ver",   opt_version_fn,"    Prints version of program" },
} ;
int   num_cmd_opts = sizeof(cmd_opts)/sizeof(cmd_opts[0]) ;


void usage ( char *progname )
{
   int   i ;
   char  fname[_MAX_FNAME];

   _splitpath( progname, NULL, NULL, fname, NULL );
   fprintf ( stderr, "Usage: %s [options] file [file]...\n", fname ) ;
   fprintf ( stderr, "Options:\n" ) ;
   for ( i = 0 ; i < num_cmd_opts ; i++ )
   {
      fprintf ( stderr, "\t%-7s%s\n", cmd_opts[i].cmd, cmd_opts[i].desc ) ;
   }
   fprintf ( stderr, "Press 'Q' to quit at any time\n" ) ;
}

int options ( int argc, char *argv[] )
{
   int   found_opt = TRUE ;
   int   i ;

   while ( found_opt )
   {
      /* force exit from the loop when we fail to recognise an option */
      found_opt = FALSE ;
      for ( i = 0 ; i < num_cmd_opts ; i++ )
      {
         if ( _strnicmp ( argv[argc],
               cmd_opts[i].cmd,
               strlen(cmd_opts[i].cmd) ) == 0 )
         {
            /* call the function associated with the command name */
            argc = cmd_opts[i].fn ( argc, argv ) ;
            found_opt = TRUE ;
         }
      }
   }

   /* rest of the command line will be treated as files */
   return argc ;
}


void main ( int argc, char *argv[] )
{
   int   i ;
   int   first_file, key = 0 ;

   /* check essential structures are correct in this compilation */
   assert ( sizeof(rgb_st) == 3 ) ;

   if ( argc < 2 )
   {
      usage ( argv[0] ) ;
      exit ( 1 ) ;
   }

   /* decode and remove the options specified */
   if ( ( first_file = options ( 1, argv ) ) == argc )
   {
      /* nothing more to do, no file names were specified */
      exit ( 0 ) ;
   }

   /* set the required screen mode */
   screen_mode ( _MRES256COLOR ) ;

   do
   {
      for ( i = first_file ; i < argc ; i++ )
      {
         last_x = 0 ;
         last_y = 0 ;                  /* setup for the slow plot routine */
         screen = _MK_FP(0xa000,0) ;   /* setup for the fast plot routine */
         decode_file ( argv[i] ) ;

         if ( frame_delay > 0 )
         {
            sleep( (clock_t)frame_delay * CLOCKS_PER_SEC ) ;
            if ( test_for_key() )
            {
               key = _getch() ;
               if ( key == 0 || key == 0xe0 ) key = _getch() ; /* get 2nd byte of 2 char keys */
               if ( key == 'Q' || key == 'q' ) i = argc ;      /* 'Q' pressed, signal exit from loop */
            }
         }
         else
         {
            while ( !test_for_key() ) ;
            key = _getch() ;
            if ( key == 0 || key == 0xe0 ) key = _getch() ; /* get 2nd byte of 2 char keys */
            if ( key == 'Q' || key == 'q' ) i = argc ;      /* 'Q' pressed, signal exit from loop */
         }
      }

      if ( key == 'Q' || key == 'q' ) break ;
   }
   while ( loop_forever ) ;

   /* back to the old screen mode */
   screen_mode ( old_mode ) ;

   exit ( 0 ) ;
}
